#include <SD.h>
#include <SPI.h>
#include <ADC.h>
#include <ADC_util.h>

// Define the array size
const int numSamples = 120000;

//ADC object
ADC *adc = new ADC();
const int readPin = A0;

// Arrays to store time and analog values
DMAMEM short timeArray[numSamples];
DMAMEM short valueArray[numSamples];

// Variables for time tracking
unsigned long previousMicros = 0;  // Last time an analog read was taken
unsigned long interval = 20;      // Interval between reads (default 100 microseconds)

int sampleIndex = 0;
bool startSampling = false;       // Flag to track when to start sampling
bool descriptionReceived = false; // Flags for sequential inputs
bool intervalReceived = false;
bool timestampReceived = false;

String descriptionText = ""; // Text input from user
String timestamp = "";       // Timestamp received from PC

File dataFile;

void setup() {
  pinMode(readPin, INPUT);
  adc->adc0->setAveraging(0); // number of averages
  adc->adc0->setResolution(10); // ADC resolution
  adc->adc0->setConversionSpeed(ADC_CONVERSION_SPEED::MED_SPEED); //conversion speed
  adc->adc0->setSamplingSpeed(ADC_SAMPLING_SPEED::MED_SPEED); // sampling speed

  Serial.begin(115200); // Initialize serial communication
  while (!Serial); // Wait for serial port to connect (for boards like Teensy 4.1)

  // Initialize the SD card
  if (!SD.begin(BUILTIN_SDCARD)) { // Assuming the SD card is connected to pin 10 (can vary)
    Serial.println("SD card initialization failed!");
    return;
  }
  Serial.println("SD card initialized.");

  // Print instructions
  Serial.println("Enter a description for the data and press ENTER.");
}

void loop() {
  // Check for serial input
  if (!startSampling && Serial.available() > 0) {
    String input = Serial.readStringUntil('\n');
    input.trim(); // Remove any extra whitespace or newline characters

    if (!descriptionReceived) {
      // First input is the description
      descriptionText = input;
      descriptionReceived = true;
      Serial.println("Description received. Enter interval (in microseconds) and press ENTER.");
    } else if (!intervalReceived) {
      // Second input is the interval
      interval = input.toInt(); // Convert input to integer
      if (interval > 0) {
        intervalReceived = true;
        Serial.println("Interval set to " + String(interval) + " microseconds. Enter timestamp in the format 'TS:yyyy-mm-dd hh:mm:ss' or type 'SKIP'.");
      } else {
        Serial.println("Invalid interval. Please enter a positive number.");
      }
    } else if (!timestampReceived) {
      // Third input is the timestamp
      if (input.equalsIgnoreCase("SKIP")) {
        timestampReceived = true;
        Serial.println("Timestamp skipped. Send 'R' to start data collection.");
      } else if (input.startsWith("TS:")) {
        timestamp = input.substring(3).replace(" ", "-").replace(":", "-");
        timestampReceived = true;
        Serial.println("Timestamp received: " + timestamp + ". Send 'R' to start data collection.");
      } else {
        Serial.println("Invalid timestamp. Please use the format 'TS:yyyy-mm-dd hh:mm:ss' or type 'SKIP'.");
      }
    } else if (input.equalsIgnoreCase("R")) {
      // 'R' starts the sampling
      startSampling = true;
      Serial.println("Data collection started.");
    } else {
      Serial.println("Invalid input. Follow the sequence: description, interval, timestamp, then 'R'.");
    }
  }

  // Sampling loop
  if (startSampling) {
    unsigned long currentMicros = micros();

    if (currentMicros - previousMicros >= interval) {
      if (sampleIndex < numSamples) {
        // Store the timestamp and analog value
        timeArray[sampleIndex] = (short)(currentMicros&0xFFFF);
        valueArray[sampleIndex] = adc->adc0->analogRead(readPin); // Fast read
        sampleIndex++;
        previousMicros = currentMicros;
      } else {
        // All samples collected, save to SD card
        delay(1000); // Allow some time before writing

        // Use timestamp or fallback filename
        String filename = (timestamp.length() > 0 ? timestamp : String(millis())) + ".csv";
        dataFile = SD.open(filename.c_str(), FILE_WRITE);

        if (dataFile) {
          // Write description, interval, and header
          dataFile.println("Description: " + descriptionText);
          dataFile.println("Interval (us): " + String(interval));
          dataFile.println("Index,Time_us,Value");

          // Write the collected data
          for (int i = 0; i < numSamples; i++) {
            dataFile.print(i);
            dataFile.print(", ");
            dataFile.print(timeArray[i]);
            dataFile.print(", ");
            dataFile.println(valueArray[i]);
            Serial.println(valueArray[i]);
          }

          // Close the file
          dataFile.close();
          Serial.println("Data written to SD card as: " + filename);
        } else {
          Serial.println("Error opening file for writing.");
        }

        // Calculate and print max/min values
        calculateAndPrintMaxMin();

        // Reset the sampling flag and inputs for another run
        Serial.println("Data collection complete. Enter a new description to start again.");
        startSampling = false;
        descriptionReceived = false;
        intervalReceived = false;
        timestampReceived = false;
        sampleIndex = 0;
      }
    }
  }
}

void calculateAndPrintMaxMin() {
  int maxValue = valueArray[0];
  int minValue = valueArray[0];

  for (int i = 1; i < numSamples; i++) {
    if (valueArray[i] > maxValue) {
      maxValue = valueArray[i];
    }
    if (valueArray[i] < minValue) {
      minValue = valueArray[i];
    }
  }

  Serial.print("Maximum Value: ");
  Serial.println(maxValue);
  Serial.print("Minimum Value: ");
  Serial.println(minValue);
}
